import asyncio
import sys
import threading

import pytest
from async_asgi_testclient import TestClient
from nextlinegraphql import create_app


@pytest.fixture(autouse=True)
def recover_trace():
    """Set the original trace function back after each test"""
    trace_org = sys.gettrace()
    yield
    sys.settrace(trace_org)
    threading.settrace(trace_org)


@pytest.fixture
async def client(app):
    async with TestClient(app) as y:
        await asyncio.sleep(0)
        yield y


@pytest.fixture
def app():
    return create_app()



import logging
from typing import cast

from alembic.migration import MigrationContext
from sqlalchemy.engine.base import Connection

from nextline_rdb import DB


def test_one(caplog):

    url = 'sqlite:///:memory:?check_same_thread=false'

    with caplog.at_level(logging.DEBUG):
        db = DB(url=url)
        engine = db.engine

    print_logrecords(caplog.records)

    with db.session() as session:
        with session.begin():
            assert session
            # print(session)

    with engine.connect() as connection:
        connection = cast(Connection, connection)
        assert connection
        # print(connection)


def print_logrecords(records):

    format = "{levelname:8s} [{name}] {message}"
    formatter = logging.Formatter(format, style="{")

    for record in records:
        message = formatter.format(record)
        # print(message)
        del message


def test_alembic():

    url = 'sqlite:///:memory:?check_same_thread=false'
    db = DB(url=url)

    with db.engine.connect() as connection:
        context = MigrationContext.configure(connection)
        rev = context.get_current_revision()
        assert rev



from typing import Type, Union

from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)


class Base(DeclarativeBase):
    metadata = metadata


class Entity(Base):
    __tablename__ = "entity"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    num: Mapped[Union[int, None]]
    txt: Mapped[Union[str, None]]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.id!r}, {self.num!r}, {self.txt!r}>"


ModelType = Type[Entity]



import contextlib
from dataclasses import dataclass, field
from logging import getLogger
from pathlib import Path
from typing import Iterator, Optional

from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import Session

from . import models


def create_tables(engine: Engine):
    '''Define tables in the database based on the ORM models .

    https://docs.sqlalchemy.org/en/20/orm/quickstart.html#emit-create-table-ddl
    '''
    models.Base.metadata.create_all(bind=engine)


@dataclass
class DB:
    '''The interface to the SQLAlchemy database.

    >>> db = DB()
    >>> with db.session() as session:
    ...     with session.begin():
    ...         pass

    '''

    url: str = 'sqlite://'
    create_engine_kwargs: dict = field(default_factory=dict)
    _engine: Optional[Engine] = field(init=False, repr=False, default=None)

    @property
    def engine(self) -> Engine:
        '''The engine with the latest alembic migration version of the table definitions.

        https://docs.sqlalchemy.org/en/20/orm/quickstart.html#create-an-engine
        '''
        if self._engine is None:
            logger = getLogger(__name__)
            logger.info(f"SQLAlchemy DB URL: {self.url}")
            self._engine = create_engine(self.url, **self.create_engine_kwargs)
            migrate_to_head(self._engine)
            create_tables(self._engine)  # NOTE: unnecessary as alembic is used
            with self._engine.connect() as connection:
                context = MigrationContext.configure(connection)
                rev = context.get_current_revision()
            logger.info(f"Alembic migration version: {rev!s}")
        return self._engine

    @contextlib.contextmanager
    def session(self) -> Iterator[Session]:
        '''A database session with session event hooks.

        The session is yielded within the outer context. The inner context, which exits
        with commit or rollback, can be used as follows:

            with db.session() as session:
                with session.begin():
                    ...

        https://docs.sqlalchemy.org/en/20/orm/session_basics.html
        '''
        with Session(self.engine) as session:
            yield session


ALEMBIC_INI = str(Path(__file__).resolve().parent / 'alembic.ini')

assert Path(ALEMBIC_INI).is_file()


def migrate_to_head(engine):
    '''Run alembic to upgrade the database to the latest version.'''
    config = Config(ALEMBIC_INI)

    # TODO: Arrange config so that logging doesn't need to be conditionally
    # configured in alembic/env.py

    # from alembic import command
    # command.upgrade(config, "head")

    # NOTE: The commented out lines of command.upgrade() above would work fine
    # for a persistent DB. Here, the following code is instead used so that the
    # migration actually takes place for an in-memory DB as well for testing.

    script = ScriptDirectory.from_config(config)

    def upgrade(rev: str, context):
        del context
        return script._upgrade_revs('head', rev)

    context = EnvironmentContext(config, script, fn=upgrade)
    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=models.Base.metadata,
            render_as_batch=True,
        )
        with context.begin_transaction():
            context.run_migrations()



from datetime import datetime
from typing import List, Type, Union

from sqlalchemy import ForeignKey, MetaData, UniqueConstraint
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

# https://docs.sqlalchemy.org/en/14/core/constraints.html#configuring-a-naming-convention-for-a-metadata-collection
# https://github.com/simonsobs/acondbs/blob/7b4e5ab967ce/acondbs/db/sa.py
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)


class Base(DeclarativeBase):
    metadata = metadata


class Hello(Base):
    __tablename__ = "hello"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    message: Mapped[Union[str, None]]


class Run(Base):
    __tablename__ = "run"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int] = mapped_column(unique=True)
    state: Mapped[Union[str, None]]
    started_at: Mapped[Union[datetime, None]]
    ended_at: Mapped[Union[datetime, None]]
    script: Mapped[Union[str, None]]
    exception: Mapped[Union[str, None]]

    traces: Mapped[List["Trace"]] = relationship(back_populates="run")
    prompts: Mapped[List["Prompt"]] = relationship(back_populates="run")
    stdouts: Mapped[List["Stdout"]] = relationship(back_populates="run")

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.run_no!r}>"


class Trace(Base):
    __tablename__ = "trace"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int]
    trace_no: Mapped[int]
    state: Mapped[str]
    thread_no: Mapped[int]
    task_no: Mapped[Union[int, None]]
    started_at: Mapped[datetime]
    ended_at: Mapped[Union[datetime, None]]

    run_id: Mapped[int] = mapped_column(ForeignKey('run.id'))
    run: Mapped[Run] = relationship(back_populates='traces')

    prompts: Mapped[List["Prompt"]] = relationship(back_populates="trace")
    stdouts: Mapped[List["Stdout"]] = relationship(back_populates="trace")

    __table_args__ = (UniqueConstraint("run_no", "trace_no"),)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.trace_no!r}>"


class Prompt(Base):
    __tablename__ = "prompt"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int]
    trace_no: Mapped[int]
    prompt_no: Mapped[int]
    open: Mapped[bool]
    event: Mapped[str]
    started_at: Mapped[datetime]
    file_name: Mapped[Union[str, None]]
    line_no: Mapped[Union[int, None]]
    stdout: Mapped[Union[str, None]]
    command: Mapped[Union[str, None]]
    ended_at: Mapped[Union[datetime, None]]

    run_id: Mapped[int] = mapped_column(ForeignKey('run.id'))
    run: Mapped[Run] = relationship(back_populates='prompts')

    trace_id: Mapped[int] = mapped_column(ForeignKey('trace.id'))
    trace: Mapped[Trace] = relationship(back_populates='prompts')

    __table_args__ = (UniqueConstraint("run_no", "prompt_no"),)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.id!r}>"


class Stdout(Base):
    __tablename__ = "stdout"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int]
    trace_no: Mapped[int]
    text: Mapped[Union[str, None]]
    written_at: Mapped[Union[datetime, None]]

    run_id: Mapped[int] = mapped_column(ForeignKey('run.id'))
    run: Mapped[Run] = relationship(back_populates='stdouts')

    trace_id: Mapped[int] = mapped_column(ForeignKey('trace.id'))
    trace: Mapped[Trace] = relationship(back_populates='stdouts')

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.text!r}>"


ModelType = Union[Type[Run], Type[Trace], Type[Prompt], Type[Stdout]]
# https://python-forum.io/thread-27697.html



from pathlib import Path

pwd = Path(__file__).resolve().parent

sub = pwd / 'queries'
QUERY_HISTORY = (sub / 'History.gql').read_text()
QUERY_HISTORY_RUNS = (sub / 'HistoryRuns.gql').read_text()

# sub = pwd / 'mutations'

# sub = pwd / 'subscriptions'



'''Run alembic check

- https://github.com/sqlalchemy/alembic/releases/tag/rel_1_9_0
- https://github.com/sqlalchemy/alembic/issues/724
- https://github.com/sqlalchemy/alembic/pull/1101
'''
from pathlib import Path

from alembic import command
from alembic.config import Config


def test_run(in_alembic_dir):
    del in_alembic_dir
    config = Config(Path.cwd() / 'alembic.ini')
    command.upgrade(config, "head")
    command.check(config)



import pytest

from nextline_rdb.pagination import SortField, load_models

from .models import Entity

params = [
    pytest.param(
        dict(sort=[SortField("num")]),
        [6, 7, 8, 9, 3, 4, 5, 0, 1, 2],
    ),
    pytest.param(
        dict(sort=[SortField("num")], after=0),
        [1, 2],
    ),
    pytest.param(
        dict(sort=[SortField("num")], before=0),
        [6, 7, 8, 9, 3, 4, 5],
    ),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_one(session, kwargs, expected):
    Model = Entity
    id_field = "id"
    models = load_models(session, Model, id_field, **kwargs)
    assert expected == [m.id for m in models]


params = [
    pytest.param(
        dict(sort=[SortField("num")]),
        ["F", "G", "H", "I", "C", "D", "E", "", "A", "B"],
    ),
    pytest.param(
        dict(sort=[SortField("num")], after=""),
        ["A", "B"],
    ),
    pytest.param(
        dict(sort=[SortField("num")], before=""),
        ["F", "G", "H", "I", "C", "D", "E"],
    ),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_str(session, kwargs, expected):
    Model = Entity
    id_field = "txt"
    models = load_models(session, Model, id_field, **kwargs)
    assert expected == [getattr(m, id_field) for m in models]


@pytest.fixture
def sample(db):
    with db.begin() as session:
        num = [3, 3, 3, 2, 2, 2, 1, 1, 1, 1]
        txt = ["", "A", "B", "C", "D", "E", "F", "G", "H", "I"]
        for i in range(10):
            model = Entity(id=i, num=num[i], txt=txt[i])
            session.add(model)



from __future__ import annotations

import strawberry
from strawberry.types import Info

from . import types
from .pagination import Connection


@strawberry.type
class History:
    runs: Connection[types.RunHistory] = strawberry.field(
        resolver=types.query_connection_run
    )
    traces: Connection[types.TraceHistory] = strawberry.field(
        resolver=types.query_connection_trace
    )
    prompts: Connection[types.PromptHistory] = strawberry.field(
        resolver=types.query_connection_prompt
    )
    stdouts: Connection[types.StdoutHistory] = strawberry.field(
        resolver=types.query_connection_stdout
    )


@strawberry.type
class Query:
    @strawberry.field
    def history(self, info: Info) -> History:
        db = info.context["db"]
        with db.session() as session:
            info.context["session"] = session
            return History()



'''Test the configuration for alembic.

Assert `nextlinegraphql/db/migration.toml` is read when the cwd is `nextlinegraphql/db`.
'''

from nextlinegraphql.config import load_settings
from nextlinegraphql.hook import load_plugins


def test_db_url(in_alembic_dir):
    del in_alembic_dir
    hook = load_plugins()
    settings = load_settings(hook=hook)
    assert settings.DB.URL == 'sqlite:///migration.sqlite3'



from sqlalchemy import select

from .models import Entity


def test_sample(session):
    Model = Entity
    stmt = select(Model)
    models = session.scalars(stmt)
    assert 10 == len(models.all())



import base64
import datetime
from typing import Any, Optional, TypedDict

import pytest
from async_asgi_testclient import TestClient
from nextlinegraphql.plugins.graphql.test import gql_request, gql_request_response

from nextline_rdb import DB
from nextline_rdb.models import Run

from ..graphql import QUERY_HISTORY_RUNS


def Cursor(i: int):
    return base64.b64encode(f"{i}".encode()).decode()


class Variables(TypedDict, total=False):
    before: str
    after: str
    first: int
    last: int


class PageInfo(TypedDict):
    hasPreviousPage: bool
    hasNextPage: bool
    startCursor: Optional[str]
    endCursor: Optional[str]


async def test_all(sample, client):
    del sample
    variables = Variables()
    expected = (False, False, Cursor(1), Cursor(100))
    await assert_results(client, variables, expected)


params = [
    pytest.param(
        Variables(first=0),
        (False, True, None, None),
        id="zero",
    ),
    pytest.param(
        Variables(first=1),
        (False, True, Cursor(1), Cursor(1)),
        id="one",
    ),
    pytest.param(
        Variables(first=5),
        (False, True, Cursor(1), Cursor(5)),
        id="fewer-than-total",
    ),
    pytest.param(
        Variables(first=99),
        (False, True, Cursor(1), Cursor(99)),
        id="fewer-than-total-by-one",
    ),
    pytest.param(
        Variables(first=100),
        (False, False, Cursor(1), Cursor(100)),
        id="exactly-total",
    ),
    pytest.param(
        Variables(first=101),
        (False, False, Cursor(1), Cursor(100)),
        id="more-than-total-by-one",
    ),
    pytest.param(
        Variables(first=150),
        (False, False, Cursor(1), Cursor(100)),
        id="more-than-total",
    ),
]


@pytest.mark.parametrize("variables, expected", params)
async def test_forward(sample, client, variables, expected):
    del sample
    await assert_results(client, variables, expected)


params = [
    pytest.param(
        Variables(after=Cursor(1)),
        (True, False, Cursor(2), Cursor(100)),
        id="one",
    ),
    pytest.param(
        Variables(after=Cursor(50)),
        (True, False, Cursor(51), Cursor(100)),
        id="middle",
    ),
    pytest.param(
        Variables(after=Cursor(99)),
        (True, False, Cursor(100), Cursor(100)),
        id="one-before-last",
    ),
    pytest.param(
        Variables(after=Cursor(100)),
        (True, False, None, None),
        id="last",
    ),
    pytest.param(
        Variables(after=Cursor(80), first=1),
        (True, True, Cursor(81), Cursor(81)),
        id="first-one",
    ),
    pytest.param(
        Variables(after=Cursor(80), first=19),
        (True, True, Cursor(81), Cursor(99)),
        id="first-fewer-than-count-by-one",
    ),
    pytest.param(
        Variables(after=Cursor(80), first=20),
        (True, False, Cursor(81), Cursor(100)),
        id="first-exactly-count",
    ),
    pytest.param(
        Variables(after=Cursor(80), first=21),
        (True, False, Cursor(81), Cursor(100)),
        id="first-more-than-count-by-one",
    ),
    pytest.param(
        Variables(after=Cursor(80), first=30),
        (True, False, Cursor(81), Cursor(100)),
        id="first-more-than-count",
    ),
    pytest.param(
        Variables(after=Cursor(100), first=0),
        (True, False, None, None),
        id="zero-after-last",
    ),
    pytest.param(
        Variables(after=Cursor(100), first=1),
        (True, False, None, None),
        id="one-after-last",
    ),
    pytest.param(
        Variables(after=Cursor(100), first=20),
        (True, False, None, None),
        id="some-after-last",
    ),
]


@pytest.mark.parametrize("variables, expected", params)
async def test_forward_with_after(sample, client, variables, expected):
    del sample
    await assert_results(client, variables, expected)


params = [
    pytest.param(
        Variables(last=0),
        (True, False, None, None),
        id="zero",
    ),
    pytest.param(
        Variables(last=1),
        (True, False, Cursor(100), Cursor(100)),
        id="one",
    ),
    pytest.param(
        Variables(last=5),
        (True, False, Cursor(96), Cursor(100)),
        id="some",
    ),
    pytest.param(
        Variables(last=99),
        (True, False, Cursor(2), Cursor(100)),
        id="fewer-than-total-by-one",
    ),
    pytest.param(
        Variables(last=100),
        (False, False, Cursor(1), Cursor(100)),
        id="exactly-total",
    ),
    pytest.param(
        Variables(last=101),
        (False, False, Cursor(1), Cursor(100)),
        id="more-than-total-by-one",
    ),
    pytest.param(
        Variables(last=150),
        (False, False, Cursor(1), Cursor(100)),
        id="more-than-total",
    ),
]


@pytest.mark.parametrize("variables, expected", params)
async def test_backward(sample, client, variables, expected):
    del sample
    await assert_results(client, variables, expected)


params = [
    pytest.param(
        Variables(before=Cursor(100)),
        (False, True, Cursor(1), Cursor(99)),
        id="end",
    ),
    pytest.param(
        Variables(before=Cursor(50)),
        (False, True, Cursor(1), Cursor(49)),
        id="middle",
    ),
    pytest.param(
        Variables(before=Cursor(2)),
        (False, True, Cursor(1), Cursor(1)),
        id="one-after-start",
    ),
    pytest.param(
        Variables(before=Cursor(1)),
        (False, True, None, None),
        id="start",
    ),
    pytest.param(
        Variables(before=Cursor(20), last=1),
        (True, True, Cursor(19), Cursor(19)),
        id="last-one",
    ),
    pytest.param(
        Variables(before=Cursor(20), last=18),
        (True, True, Cursor(2), Cursor(19)),
        id="last-fewer-than-count-by-one",
    ),
    pytest.param(
        Variables(before=Cursor(20), last=19),
        (False, True, Cursor(1), Cursor(19)),
        id="last-exactly-count",
    ),
    pytest.param(
        Variables(before=Cursor(20), last=20),
        (False, True, Cursor(1), Cursor(19)),
        id="last-more-than-count-by-one",
    ),
    pytest.param(
        Variables(before=Cursor(20), last=30),
        (False, True, Cursor(1), Cursor(19)),
        id="last-more-than-count",
    ),
]


@pytest.mark.parametrize("variables, expected", params)
async def test_backward_with_before(sample, client, variables, expected):
    del sample
    await assert_results(client, variables, expected)


params = [
    pytest.param(
        Variables(),
        (False, False, Cursor(1), Cursor(1)),
        id="default",
    ),
    pytest.param(
        Variables(after=Cursor(1)),
        (True, False, None, None),
        id="forward-after-one",
    ),
    pytest.param(
        Variables(first=1),
        (False, False, Cursor(1), Cursor(1)),
        id="forward-first-one",
    ),
    pytest.param(
        Variables(after=Cursor(1), first=1),
        (True, False, None, None),
        id="forward-after-one-first-one",
    ),
    pytest.param(
        Variables(before=Cursor(1)),
        (False, True, None, None),
        id="backward-before-one",
    ),
    pytest.param(
        Variables(last=1),
        (False, False, Cursor(1), Cursor(1)),
        id="backward-last-one",
    ),
    pytest.param(
        Variables(before=Cursor(1), last=1),
        (False, True, None, None),
        id="backward-before-one-last-one",
    ),
]


@pytest.mark.parametrize("variables, expected", params)
async def test_one(sample_one, client, variables, expected):
    del sample_one
    await assert_results(client, variables, expected)


@pytest.mark.parametrize("first_or_last", ["first", "last"])
@pytest.mark.parametrize("number", [None, 0, 1, 5])
async def test_empty(sample_empty, client, first_or_last, number):
    del sample_empty
    variables = {}
    if number is not None:
        variables[first_or_last] = number
    expected = (False, False, None, None)
    await assert_results(client, variables, expected)


async def assert_results(client: TestClient, variables, expected):

    expected_page_info = PageInfo(
        hasPreviousPage=expected[0],
        hasNextPage=expected[1],
        startCursor=expected[2],
        endCursor=expected[3],
    )

    data = await gql_request(client, QUERY_HISTORY_RUNS, variables=variables)

    all_runs = data["history"]["runs"]
    page_info = all_runs["pageInfo"]
    edges = all_runs["edges"]

    # print(page_info)
    print(edges)

    assert expected_page_info == page_info

    if start_cursor := expected_page_info["startCursor"]:
        edge = edges[0]
        assert start_cursor == edge["cursor"]
        assert start_cursor == Cursor(edge["node"]["id"])
    else:
        assert not edges

    if end_cursor := expected_page_info["endCursor"]:
        edge = edges[-1]
        assert end_cursor == edge["cursor"]
        assert end_cursor == Cursor(edge["node"]["id"])
    else:
        assert not edges


params = [
    pytest.param(
        Variables(first=5, last=5),
        id="first-and-last",
    ),
    pytest.param(
        Variables(before=Cursor(31), first=5),
        id="before-and-first",
    ),
    pytest.param(
        Variables(after=Cursor(20), last=5),
        id="after-and-last",
    ),
    pytest.param(
        Variables(before=Cursor(31), after=Cursor(20)),
        id="before-and-after",
    ),
    pytest.param(
        Variables(before=Cursor(31), after=Cursor(20), first=5, last=5),
        id="all",
    ),
]


@pytest.mark.parametrize("variables", params)
async def test_error_forward_and_backward(sample, client, variables):
    del sample

    resp = await gql_request_response(
        client,
        QUERY_HISTORY_RUNS,
        variables=variables,
    )
    result = resp.json()
    assert result["data"] is None
    assert result["errors"]


@pytest.fixture
def sample(db: DB):
    with db.session() as session:
        with session.begin():
            for run_no in range(11, 111):
                model = Run(
                    run_no=run_no,
                    state="running",
                    started_at=datetime.datetime.utcnow(),
                    ended_at=datetime.datetime.utcnow(),
                    script="pass",
                )
                session.add(model)


@pytest.fixture
def sample_one(db: DB):
    with db.session() as session:
        with session.begin():
            run_no = 10
            model = Run(
                run_no=run_no,
                state="running",
                started_at=datetime.datetime.utcnow(),
                ended_at=datetime.datetime.utcnow(),
                script="pass",
            )
            session.add(model)


@pytest.fixture
def sample_empty(db: DB):
    del db


@pytest.fixture
def app(db: DB):
    # NOTE: Overriding the app fixture from conftest.py because it adds an
    # entry in the DB. The factory.create_app() needs to be refactored so this
    # override is not needed.
    import strawberry
    from nextlinegraphql.custom.strawberry import GraphQL
    from starlette.applications import Starlette

    from nextline_rdb.schema import Query

    schema = strawberry.Schema(query=Query)

    class EGraphQL(GraphQL):
        """Extend the strawberry GraphQL app

        https://strawberry.rocks/docs/integrations/asgi
        """

        async def get_context(self, request, response=None) -> Optional[Any]:
            return {"request": request, "response": response, "db": db}

    app_ = EGraphQL(schema)

    ret = Starlette(debug=True)
    ret.mount("/", app_)
    return ret


@pytest.fixture
def db():
    url = 'sqlite:///:memory:?check_same_thread=false'
    return DB(url=url)



from __future__ import annotations

import asyncio
from collections import deque
from contextlib import asynccontextmanager
from logging import getLogger
from typing import TYPE_CHECKING, Deque, List, Union

from sqlalchemy import select

from . import models as db_models
from .db import DB

if TYPE_CHECKING:
    from nextline import Nextline
    from nextline.types import StdoutInfo


@asynccontextmanager
async def write_db(nextline: Nextline, db: DB, timeout_on_exit: float = 3):
    task = asyncio.gather(
        subscribe_run_info(nextline, db),
        subscribe_trace_info(nextline, db),
        subscribe_prompt_info(nextline, db),
        subscribe_stdout(nextline, db),
    )
    try:
        yield
    finally:
        try:
            await asyncio.wait_for(task, timeout=timeout_on_exit)
        except BaseException:
            logger = getLogger(__name__)
            logger.exception("Exception occurred in writing to DB")
            raise


async def subscribe_run_info(nextline: Nextline, db: DB):
    async for run_info in nextline.subscribe_run_info():
        run_no = run_info.run_no
        with db.session() as session:
            with session.begin():
                model: Union[db_models.Run, None]
                if run_info.state == "initialized":
                    model = db_models.Run(
                        run_no=run_no,
                        state=run_info.state,
                        script=run_info.script,
                    )
                    session.add(model)
                elif run_info.state == "running":
                    stmt = select(db_models.Run).filter_by(run_no=run_no)
                    while not (model := session.execute(stmt).scalar_one_or_none()):
                        await asyncio.sleep(0)
                    model.state = run_info.state
                    model.started_at = run_info.started_at
                elif run_info.state == "finished":
                    stmt = select(db_models.Run).filter_by(run_no=run_no)
                    while not (model := session.execute(stmt).scalar_one_or_none()):
                        await asyncio.sleep(0)
                    model.state = run_info.state
                    model.ended_at = run_info.ended_at
                    model.exception = run_info.exception


async def subscribe_trace_info(nextline: Nextline, db: DB):
    async for trace_info in nextline.subscribe_trace_info():
        with db.session() as session:
            with session.begin():
                model: Union[db_models.Trace, None]
                stmt_runs = select(db_models.Run).filter_by(run_no=trace_info.run_no)
                while not (run := session.execute(stmt_runs).scalar_one_or_none()):
                    await asyncio.sleep(0)
                if trace_info.state == "running":
                    model = db_models.Trace(
                        run_no=trace_info.run_no,
                        trace_no=trace_info.trace_no,
                        state=trace_info.state,
                        thread_no=trace_info.thread_no,
                        task_no=trace_info.task_no,
                        started_at=trace_info.started_at,
                        run=run,
                    )
                    session.add(model)
                elif trace_info.state == "finished":
                    stmt_traces = select(db_models.Trace).filter_by(
                        run_no=trace_info.run_no,
                        trace_no=trace_info.trace_no,
                    )
                    while not (
                        model := session.execute(stmt_traces).scalar_one_or_none()
                    ):
                        await asyncio.sleep(0)
                    model.state = trace_info.state
                    model.ended_at = trace_info.ended_at


async def subscribe_prompt_info(nextline: Nextline, db: DB):
    async for prompt_info in nextline.subscribe_prompt_info():
        if prompt_info.trace_call_end:  # TODO: remove when unnecessary
            continue
        with db.session() as session:
            with session.begin():
                model: Union[db_models.Prompt, None]
                stmt_runs = select(db_models.Run).filter_by(run_no=prompt_info.run_no)
                while not (run := session.execute(stmt_runs).scalar_one_or_none()):
                    await asyncio.sleep(0)
                stmt_traces = select(db_models.Trace).filter_by(
                    run_no=prompt_info.run_no, trace_no=prompt_info.trace_no
                )
                while not (trace := session.execute(stmt_traces).scalar_one_or_none()):
                    await asyncio.sleep(0)
                if prompt_info.open:
                    model = db_models.Prompt(
                        run_no=prompt_info.run_no,
                        trace_no=prompt_info.trace_no,
                        prompt_no=prompt_info.prompt_no,
                        open=prompt_info.open,
                        event=prompt_info.event,
                        file_name=prompt_info.file_name,
                        line_no=prompt_info.line_no,
                        stdout=prompt_info.stdout,
                        started_at=prompt_info.started_at,
                        run=run,
                        trace=trace,
                    )
                    session.add(model)
                else:
                    stmt_prompts = select(db_models.Prompt).filter_by(
                        run_no=prompt_info.run_no,
                        prompt_no=prompt_info.prompt_no,
                    )
                    while not (
                        model := session.execute(stmt_prompts).scalar_one_or_none()
                    ):
                        await asyncio.sleep(0)
                    model.open = prompt_info.open
                    model.command = prompt_info.command
                    model.ended_at = prompt_info.ended_at


async def subscribe_stdout(nextline: Nextline, db: DB):

    run_info = None
    stdout_info_list: Deque[StdoutInfo] = deque()
    lock = asyncio.Condition()

    async def f():
        nonlocal stdout_info_list
        async for s in nextline.subscribe_stdout():
            # print(s, file=sys.stderr)
            async with lock:
                stdout_info_list.append(s)

    t = asyncio.create_task(f())

    async for run_info in nextline.subscribe_run_info():
        if not run_info.state == "finished":
            continue
        async with lock:
            to_save: List[StdoutInfo] = []
            while stdout_info_list:
                info = stdout_info_list.popleft()
                if info.run_no < run_info.run_no:
                    continue
                if run_info.run_no < info.run_no:
                    stdout_info_list.appendleft(info)
                    break
                to_save.append(info)
        with db.session() as session:
            with session.begin():
                for info in to_save:
                    stmt_runs = select(db_models.Run).filter_by(run_no=info.run_no)
                    while not (run := session.execute(stmt_runs).scalar_one_or_none()):
                        await asyncio.sleep(0)
                    stmt_traces = select(db_models.Trace).filter_by(
                        run_no=info.run_no, trace_no=info.trace_no
                    )
                    while not (
                        trace := session.execute(stmt_traces).scalar_one_or_none()
                    ):
                        await asyncio.sleep(0)
                    model = db_models.Stdout(
                        run_no=info.run_no,
                        trace_no=info.trace_no,
                        text=info.text,
                        written_at=info.written_at,
                        run=run,
                        trace=trace,
                    )
                    session.add(model)

    t.cancel()
